/*
 *  Copyright (c) 2016 Intel Corporation.
 *  Intel Corporation All Rights Reserved.
 *
 *  Portions of the source code contained or described herein and all documents related
 *  to portions of the source code ("Material") are owned by Intel Corporation or its
 *  suppliers or licensors.  Title to the Material remains with Intel
 *  Corporation or its suppliers and licensors.  The Material contains trade
 *  secrets and proprietary and confidential information of Intel or its
 *  suppliers and licensors.  The Material is protected by worldwide copyright
 *  and trade secret laws and treaty provisions.  No part of the Material may
 *  be used, copied, reproduced, modified, published, uploaded, posted,
 *  transmitted, distributed, or disclosed in any way without Intel's prior
 *  express written permission.
 *
 *  No license under any patent, copyright, trade secret or other intellectual
 *  property right is granted to or conferred upon you by disclosure or
 *  delivery of the Materials, either expressly, by implication, inducement,
 *  estoppel or otherwise. Any license under such intellectual property rights
 *  must be express and approved by Intel in writing.
 *
 */ 
#include <hStreams_app_api.h> 
#include <stdio.h>
#include <omp.h>
#include <math.h>
#include <mpi.h>
#include <unistd.h>
#define MASTER 0

const long ARRAY_SIZE = 1024*1024; 
const int NUM_DOMS = 2;
static HSTR_OPTIONS hstreams_options;

void rand_initialize(double *X, long int n)
{
    #pragma omp parallel for
    for (long int i = 0; i < n; ++i) {
        X[i] = rand() % 100 + 0.1;
    }
}

int main(int argc, char *argv[])
{
    int mpi_id, num_procs, namelen;
    char name[MPI_MAX_PROCESSOR_NAME];
    MPI_Status stat;
    const int TAG_DISTRIBUTE_A0 = 1;
    const int TAG_DISTRIBUTE_A1 = 2;
    const int TAG_DISTRIBUTE_A2 = 3;
    const int TAG_DISTRIBUTE_B0 = 4;
    const int TAG_DISTRIBUTE_B1 = 5;
    const int TAG_DISTRIBUTE_B2 = 6;
    const int TAG_DISTRIBUTE_C0 = 7;
    const int TAG_DISTRIBUTE_C1 = 8;
    const int TAG_DISTRIBUTE_C2 = 9;

    if (MPI_Init (&argc, &argv) != MPI_SUCCESS)
    {
       printf ("Failed to initialize MPI\n");
       return (-1);
    }

    MPI_Comm_size (MPI_COMM_WORLD, &num_procs);

    if (num_procs != 2)
    {
       printf ("Please run this program with 2 ranks!\n");
       return(-2);
    }

    MPI_Comm_rank (MPI_COMM_WORLD, &mpi_id);
    MPI_Get_processor_name (name, &namelen); 

    printf("Hello world! rank %d of %d runs on %s\n", mpi_id, num_procs, name);  
    double diff;
    int streams_per_domain = 1; // 1 stream per card
    // no over-subscription, i.e., one logical stream per physical stream.
    int oversubscription = 1;
    CHECK_HSTR_RESULT(hStreams_app_init(streams_per_domain, oversubscription));

    long int NN = ARRAY_SIZE/2; 
    long int N = NN/2;
    double *A0, *B0, *C0;
    double *A1, *B1, *C1;

    A0 = new double[N];
    B0 = new double[N];
    C0 = new double[N];
    A1 = new double[N];
    B1 = new double[N];
    C1 = new double[N];
    
    double *AA, *BB, *CC;
    if (mpi_id == MASTER)
    {
       AA = new double[ARRAY_SIZE];
       BB = new double[ARRAY_SIZE];
       CC = new double[ARRAY_SIZE];

       srand(0123);
       rand_initialize(AA, ARRAY_SIZE);
       rand_initialize(BB, ARRAY_SIZE);

       memcpy(A0, AA, N*sizeof(double));
       memcpy(B0, BB, N*sizeof(double));
       memcpy(A1, AA+N, N*sizeof(double));
       memcpy(B1, BB+N, N*sizeof(double));
 
       MPI_Send(AA+NN, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_A0, MPI_COMM_WORLD);
       MPI_Send(BB+NN, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_B0, MPI_COMM_WORLD);
       MPI_Send(AA+NN+N, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_A1, MPI_COMM_WORLD);
       MPI_Send(BB+NN+N, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_B1, MPI_COMM_WORLD);
    }   
    else
    {
       MPI_Recv(A0, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_A0, MPI_COMM_WORLD, &stat);
       MPI_Recv(B0, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_B0, MPI_COMM_WORLD, &stat);
       MPI_Recv(A1, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_A1, MPI_COMM_WORLD, &stat);
       MPI_Recv(B1, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_B1, MPI_COMM_WORLD, &stat);
    } 

    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)A0, N * sizeof(double)));
    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)B0, N * sizeof(double)));
    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)C0, N * sizeof(double)));
    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)A1, N * sizeof(double)));
    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)B1, N * sizeof(double)));
    CHECK_HSTR_RESULT(hStreams_app_create_buf((void *)C1, N * sizeof(double)));
    
    HSTR_EVENT completionA[NUM_DOMS], completionB[NUM_DOMS], completionC[NUM_DOMS], completionD[NUM_DOMS];
    double *pA[NUM_DOMS], *pB[NUM_DOMS], *pC[NUM_DOMS];
    
    pA[0] = A0;
    pB[0] = B0;
    pC[0] = C0;
    pA[1] = A1;
    pB[1] = B1;
    pC[1] = C1;
    
    int str_id;
    uint64_t args2[5];
    
    for (str_id=0; str_id<NUM_DOMS; str_id++)
    {
       printf("Rank %d: stream %d moves A\n", mpi_id, str_id);
       CHECK_HSTR_RESULT(hStreams_app_xfer_memory(
                      str_id, 
                      pA[str_id],
                      pA[str_id], 
                      N * sizeof(double),
                      HSTR_SRC_TO_SINK,
                      &completionA[str_id]));

       printf("Rank %d: stream %d moves B\n", mpi_id, str_id);
       CHECK_HSTR_RESULT(hStreams_app_xfer_memory(
                      str_id,
                      pB[str_id],
                      pB[str_id], 
                      N * sizeof(double), 
                      HSTR_SRC_TO_SINK, 
                      &completionB[str_id]));
    }

    for (str_id=0; str_id<NUM_DOMS; str_id++)
    {
       args2[0] = (uint64_t) N;
       args2[1] = (uint64_t) str_id;

       if (str_id == 0)
       {
          args2[2] = (uint64_t) A0;
          args2[3] = (uint64_t) B0;
          args2[4] = (uint64_t) C0;
       } else if (str_id == 1) 
       {
          args2[2] = (uint64_t) A1;
          args2[3] = (uint64_t) B1;
          args2[4] = (uint64_t) C1;
       } 

       printf("Rank %d: compute on stream %d\n", mpi_id, str_id);

       int j;
       for (j=0;j<1;j++)
       CHECK_HSTR_RESULT(hStreams_app_invoke(
                      str_id,
                      "vector_add",
                      2,
                      3,
                      args2,
                      &completionC[str_id],
                      NULL, 
                      0));
    }
   
    for (str_id=0; str_id<NUM_DOMS; str_id++)
    {
       printf("Rank %d: stream %d Xtransfer data in C back\n", mpi_id, str_id);
       CHECK_HSTR_RESULT(hStreams_app_xfer_memory(
                      str_id,
                      pC[str_id],
                      pC[str_id], 
                      N * sizeof(double), 
                      HSTR_SINK_TO_SRC,
                      &completionD[str_id])); 
       
       CHECK_HSTR_RESULT(hStreams_app_event_wait(
                      1, 
                      &completionD[str_id])); 
    }

    CHECK_HSTR_RESULT(hStreams_app_fini());
   
    printf("C0=%4.2f ",*C0);
    printf("C1=%4.2f ",*C1);
 
    if (mpi_id == MASTER)
    {
       memcpy(CC, C0, N*sizeof(double));
       memcpy(CC+N, C1, N*sizeof(double));

       MPI_Recv(CC+NN, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_C0, MPI_COMM_WORLD, &stat);
       MPI_Recv(CC+NN+N, N, MPI_DOUBLE, 1, TAG_DISTRIBUTE_C1, MPI_COMM_WORLD, &stat);
    }    
    else
    {
       MPI_Send(C0, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_C0, MPI_COMM_WORLD);
       MPI_Send(C1, N, MPI_DOUBLE, MASTER, TAG_DISTRIBUTE_C1, MPI_COMM_WORLD);
    } 
 
    if (mpi_id == MASTER)
    {
       int pass = 1;
       for (long int i=0; i<ARRAY_SIZE; i++) 
       {
          diff = AA[i] + BB[i] - CC[i]; 
	  if (fabs(diff) > 0.00001)
          {
             pass = 0;
             printf("******AA[%ld]=%4.2f  BB[%ld]=%4.2f CC[%ld]=%4.2f PASS:%d\n", \
                  i, AA[i], i, BB[i], i, CC[i], pass);
          }
       }

       if (pass == 1) 
          printf("PASSED!\n");
       else
          printf("FAILED!\n");
    }

    delete []A0;
    delete []B0;
    delete []C0;
    delete []A1;
    delete []B1;
    delete []C1;

    if (mpi_id == MASTER)
    {
       delete []AA;
       delete []BB;
       delete []CC;
    }

    MPI_Finalize();
    return 0;
}
